'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

function _objectWithoutPropertiesLoose(source, excluded) {
  if (source == null) return {};
  var target = {};
  var sourceKeys = Object.keys(source);
  var key, i;

  for (i = 0; i < sourceKeys.length; i++) {
    key = sourceKeys[i];
    if (excluded.indexOf(key) >= 0) continue;
    target[key] = source[key];
  }

  return target;
}

var isString = function isString(obj) {
  return typeof obj === 'string';
};
var isFunction = function isFunction(obj) {
  return typeof obj === 'function';
};
var validateName = function validateName(name) {
  if (!isString(name)) {
    throw new Error("Name must be a string, but recevied " + typeof name);
  }
};
var validateDependencies = function validateDependencies(dependencies) {
  if (!Array.isArray(dependencies)) {
    throw new Error("Expected an array of dependencies, but was called with [" + typeof dependencies + "]");
  }
};
var validateFactory = function validateFactory(name, fn) {
  if (!isFunction(fn)) {
    throw new Error("Factory " + name + " must be a [Function]");
  }
};
var validateArgs = function validateArgs(name, dependencies, fn) {
  validateName(name);
  validateDependencies(dependencies);
  validateFactory(name, fn);
};
var isPassive = function isPassive(name, jpex, precedence) {
  return (precedence || jpex.$$config.precedence) === 'passive' && jpex.$$factories[name] != null;
};
var instantiate = function instantiate(context, args) {
  return new (Function.prototype.bind.apply(context, args))();
};
var isNode = function isNode() {
  var _process;
  try {
    _process = new Function('return process')();
  } catch (e) {
  }
  return typeof _process === 'object' && _process.toString && _process.toString() === '[object process]';
};
var doUnsafeRequire = new Function('require', 'target', 'return require.main.require(target)');
var unsafeRequire = function unsafeRequire(target) {
  return doUnsafeRequire(eval('require'), target);
};
var ensureArray = function ensureArray(arr) {
  if (arr == null) {
    return [];
  }
  if (Array.isArray(arr)) {
    return arr;
  }
  return [arr];
};
var hasLength = function hasLength(arr) {
  return arr == null || arr.length > 0;
};

function constant(name, obj) {
  validateName(name);
  if (isPassive(name, this)) {
    return;
  }
  this.$$factories[name] = {
    fn: function fn() {
      return obj;
    },
    lifecycle: 'application',
    value: obj,
    resolved: true
  };
}

function factory(name, dependencies, fn, opts) {
  var _this = this;
  if (opts === void 0) {
    opts = {};
  }
  validateArgs(name, dependencies, fn);
  if (!hasLength(dependencies)) {
    dependencies = null;
  }
  if (isPassive(name, this, opts.precedence)) {
    return;
  }
  this.$$factories[name] = {
    fn: fn,
    dependencies: dependencies,
    lifecycle: opts.lifecycle || this.$$config.lifecycle
  };
  if (opts.alias) {
    ensureArray(opts.alias).forEach(function (alias) {
      return _this.alias(alias, name);
    });
  }
}

function service(name, dependencies, fn, opts) {
  if (opts === void 0) {
    opts = {};
  }
  validateArgs(name, dependencies, fn);
  function factory() {
    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }
    var context = {};
    if (opts.bindToInstance) {
      dependencies.forEach(function (key, i) {
        context[key] = args[i];
      });
    }
    return instantiate(fn, [context].concat(args));
  }
  return this.factory(name, dependencies, factory, opts);
}

function alias(alias, name) {
  if (this.$$alias[alias] == null || this.$$config.precedence === 'active') {
    this.$$alias[alias] = name;
  }
  if (this.$$alias[name] == null || this.$$config.precedence === 'active') {
    this.$$alias[name] = alias;
  }
}

var GLOBAL_TYPE_PREFIX = 'type:global:';
var NAMED_PARAMS = '$namedParameters';
var VOID = 'undefined';

var getFromNodeModules = function getFromNodeModules(jpex, target) {
  if (!jpex.$$config.nodeModules || !isNode()) {
    return;
  }
  try {
    var value = unsafeRequire(target);
    jpex.constant(target, value);
    return jpex.$$factories[target];
  } catch (e) {
    var _e$message;
    if ((_e$message = e.message) != null && _e$message.includes != null && _e$message.includes("Cannot find module '" + target + "'")) {
      return;
    }
    throw e;
  }
};
var getGlobalObject = function getGlobalObject() {
  if (typeof global !== VOID) {
    return global;
  }
  if (typeof globalThis !== VOID) {
    return globalThis;
  }
  if (typeof window !== VOID) {
    return window;
  }
  return {};
};
var getGlobalProperty = function getGlobalProperty(name) {
  var global = getGlobalObject();
  if (global[name] !== void 0) {
    return global[name];
  }
  if (name.startsWith(GLOBAL_TYPE_PREFIX)) {
    var _global$inferredLower;
    var len = GLOBAL_TYPE_PREFIX.length;
    var inferred = name.substr(len);
    var inferredLower = inferred.charAt(0).toLowerCase() + inferred.substr(1);
    return (_global$inferredLower = global[inferredLower]) != null ? _global$inferredLower : global[inferred];
  }
};
var getFromGlobal = function getFromGlobal(jpex, name) {
  if (!jpex.$$config.globals) {
    return;
  }
  var value = getGlobalProperty(name);
  if (value !== void 0) {
    jpex.constant(name, value);
    return jpex.$$factories[name];
  }
};
var getFromAlias = function getFromAlias(jpex, alias) {
  var name = jpex.$$alias[alias];
  if (name != null) {
    return jpex.$$factories[name];
  }
};
var getFromResolved = function getFromResolved(jpex, name) {
  return jpex.$$resolved[name];
};
var getFromRegistry = function getFromRegistry(jpex, name) {
  return jpex.$$factories[name];
};
var getFactory = function getFactory(jpex, name, opts) {
  var _opts$optional;
  if (opts === void 0) {
    opts = {};
  }
  validateName(name);
  var fns = [getFromResolved, getFromRegistry, getFromAlias, getFromGlobal, getFromNodeModules];
  while (fns.length) {
    var factory = fns.shift()(jpex, name);
    if (factory != null) {
      return factory;
    }
  }
  if ((_opts$optional = opts.optional) != null ? _opts$optional : jpex.$$config.optional) {
    return;
  }
  throw new Error("Unable to find required dependency [" + name + "]");
};
var cacheResult = function cacheResult(jpex, name, factory, value, namedParameters, withArg) {
  switch (factory.lifecycle) {
    case 'application':
      factory.resolved = true;
      factory.value = value;
      factory.with = withArg;
      break;
    case 'class':
      jpex.$$resolved[name] = _extends({}, factory, {
        resolved: true,
        value: value,
        with: withArg
      });
      break;
    case 'none':
      break;
    case 'instance':
    default:
      namedParameters[name] = value;
      break;
  }
};
var checkStack = function checkStack(jpex, name, stack) {
  if (!hasLength(stack)) {
    return false;
  }
  if (!stack.includes(name)) {
    return false;
  }
  if (stack[stack.length - 1] === name) {
    var _jpex$$$parent;
    var parent = (_jpex$$$parent = jpex.$$parent) == null ? void 0 : _jpex$$$parent.$$factories[name];
    if (parent != null) {
      return true;
    }
  }
  throw new Error("Recursive loop for dependency " + name + " encountered");
};

var getNamedParameters = function getNamedParameters(namedParameters, opts) {
  if (opts === void 0) {
    opts = {};
  }
  if (namedParameters) {
    return namedParameters;
  }
  if (opts.with) {
    return _extends({}, opts.with);
  }
  return {};
};
var isResolvedWithParams = function isResolvedWithParams(factory, opts) {
  if (opts === void 0) {
    opts = {};
  }
  if (!factory.with && !opts.with) {
    return true;
  }
  var keys = [].concat(new Set([].concat(Object.keys(opts.with || {}), Object.keys(factory.with || {}))));
  return keys.every(function (key) {
    var _opts$with, _factory$with;
    return ((_opts$with = opts.with) == null ? void 0 : _opts$with[key]) === ((_factory$with = factory.with) == null ? void 0 : _factory$with[key]);
  });
};
var resolveFactory = function resolveFactory(jpex, name, factory, namedParameters, opts, stack) {
  if (factory == null) {
    return;
  }
  if (factory.resolved && isResolvedWithParams(factory, opts)) {
    return factory.value;
  }
  var args = [];
  if (hasLength(factory.dependencies)) {
    args = resolveMany(jpex, factory, namedParameters, opts, [].concat(stack, [name]));
  }
  var value = factory.fn.apply(jpex, args);
  cacheResult(jpex, name, factory, value, namedParameters, opts == null ? void 0 : opts.with);
  return value;
};
var resolveOne = function resolveOne(jpex, name, initialParameters, opts, stack) {
  var namedParameters = getNamedParameters(initialParameters, opts);
  if (Object.hasOwnProperty.call(namedParameters, name)) {
    return namedParameters[name];
  }
  if (name === NAMED_PARAMS || name === "type:jpex/NamedParameters") {
    return namedParameters;
  }
  if (checkStack(jpex, name, stack)) {
    return resolveOne(jpex.$$parent, name, namedParameters, opts, []);
  }
  var factory = getFactory(jpex, name, opts);
  return resolveFactory(jpex, name, factory, namedParameters, opts, stack);
};
var resolveMany = function resolveMany(jpex, definition, namedParameters, opts, stack) {
  if (stack === void 0) {
    stack = [];
  }
  if (!hasLength(definition.dependencies)) {
    return [];
  }
  var dependencies = ensureArray(definition.dependencies);
  var values = dependencies.map(function (dependency) {
    return resolveOne(jpex, dependency, namedParameters, opts, stack);
  });
  return values;
};

function resolve(name, opts) {
  return resolveOne(this, name, void 0, opts, []);
}
function resolveDependencies(definition, opts) {
  return resolveMany(this, definition, void 0, opts, []);
}
function isResolved(dependency) {
  var _this$$$factories$dep;
  if (!isString(dependency)) {
    return false;
  }
  if (this.$$resolved[dependency] != null) {
    return true;
  }
  return ((_this$$$factories$dep = this.$$factories[dependency]) == null ? void 0 : _this$$$factories$dep.resolved) === true;
}
function allResolved(dependencies) {
  return dependencies.every(isResolved.bind(this));
}

function encase(dependencies, fn) {
  var jpex = this;
  var result;
  var encased = function encased() {
    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }
    if (result && allResolved.call(jpex, dependencies)) {
      return result.apply(this, args);
    }
    var deps = resolveDependencies.call(jpex, {
      dependencies: dependencies
    });
    result = fn.apply(jpex, deps);
    return result.apply(this, args);
  };
  encased.encased = fn;
  return encased;
}

function clearCache() {
  for (var _len = arguments.length, names = new Array(_len), _key = 0; _key < _len; _key++) {
    names[_key] = arguments[_key];
  }
  names = ensureArray(names);
  for (var key in this.$$factories) {
    if (!hasLength(names) || names.includes(key)) {
      this.$$factories[key].resolved = false;
    }
  }
  for (var _key2 in this.$$resolved) {
    if (!hasLength(names) || names.includes(_key2)) {
      delete this.$$resolved[_key2];
    }
  }
}

var defaultConfig = {
  lifecycle: 'class',
  precedence: 'active',
  globals: true,
  nodeModules: true,
  optional: false
};
function makeJpex(_temp, parent) {
  var _ref = _temp === void 0 ? {} : _temp,
      _ref$inherit = _ref.inherit,
      inherit = _ref$inherit === void 0 ? true : _ref$inherit,
      config = _objectWithoutPropertiesLoose(_ref, ["inherit"]);
  var jpex = {
    $$parent: parent,
    $$config: _extends({}, defaultConfig, inherit ? parent == null ? void 0 : parent.$$config : {}, config),
    $$factories: parent && inherit ? Object.create(parent.$$factories) : {},
    $$resolved: {},
    $$alias: parent && inherit ? Object.create(parent.$$alias) : {},
    constant: constant,
    factory: factory,
    service: service,
    alias: alias,
    resolve: resolve,
    encase: encase,
    clearCache: clearCache,
    extend: function extend(config) {
      return makeJpex(config, this);
    },
    resolveWith: function resolveWith(name, namedParameters, opts) {
      return this.resolve(name, _extends({
        with: namedParameters
      }, opts));
    },
    raw: function raw(name) {
      return getFactory(this, name, {}).fn;
    },
    infer: function infer() {
      return '';
    }
  };
  return jpex;
}

var jpex = makeJpex();

exports.default = jpex;
exports.jpex = jpex;
